package port

import (
	"net/http"
	"strconv"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/port/internal/apimodel"
)

const apiPlayerSnapshotSortMaxLength = 1

var apiPlayerSnapshotSortAllowedValues = []domain.PlayerSnapshotSort{
	domain.PlayerSnapshotSortDateASC,
	domain.PlayerSnapshotSortDateDESC,
}

//nolint:gocyclo
func (h *apiHTTPHandler) ListPlayerPlayerSnapshots(
	w http.ResponseWriter,
	r *http.Request,
	_ apimodel.VersionCodePathParam,
	serverKey apimodel.ServerKeyPathParam,
	playerID apimodel.PlayerIdPathParam,
	params apimodel.ListPlayerPlayerSnapshotsParams,
) {
	domainParams := domain.NewListPlayerSnapshotsParams()

	if err := domainParams.SetSort([]domain.PlayerSnapshotSort{domain.PlayerSnapshotSortIDASC}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
		return
	}

	if err := domainParams.SetPlayerIDs([]int{playerID}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
		return
	}

	if params.Sort != nil {
		if err := domainParams.PrependSortString(
			*params.Sort,
			apiPlayerSnapshotSortAllowedValues,
			apiPlayerSnapshotSortMaxLength,
		); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
			return
		}
	} else {
		if err := domainParams.PrependSort([]domain.PlayerSnapshotSort{domain.PlayerSnapshotSortDateASC}); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
			return
		}
	}

	if err := domainParams.SetServerKeys([]string{serverKey}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
		return
	}

	if params.Limit != nil {
		if err := domainParams.SetLimit(*params.Limit); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
			return
		}
	}

	if params.Cursor != nil {
		if err := domainParams.SetEncodedCursor(*params.Cursor); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayerSnapshotsErrorPath).render(w, r, err)
			return
		}
	}

	res, err := h.playerSnapshotSvc.ListWithRelations(r.Context(), domainParams)
	if err != nil {
		h.errorRenderer.render(w, r, err)
		return
	}

	renderJSON(w, r, http.StatusOK, apimodel.NewListPlayerSnapshotsResponse(res))
}

func formatListPlayerSnapshotsErrorPath(segments []domain.ErrorPathSegment) []string {
	if segments[0].Model != "ListPlayerSnapshotsParams" {
		return nil
	}

	switch segments[0].Field {
	case "cursor":
		return []string{"$query", "cursor"}
	case "limit":
		return []string{"$query", "limit"}
	case "sort":
		path := []string{"$query", "sort"}
		if segments[0].Index >= 0 {
			path = append(path, strconv.Itoa(segments[0].Index))
		}
		return path
	default:
		return nil
	}
}
