package port_test

import (
	"cmp"
	"fmt"
	"net/http"
	"net/http/httptest"
	"slices"
	"strconv"
	"strings"
	"testing"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/domain/domaintest"
	"gitea.dwysokinski.me/twhelp/corev3/internal/port/internal/apimodel"
	"github.com/brianvoe/gofakeit/v7"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

const endpointListPlayerPlayerSnapshots = "/v2/versions/%s/servers/%s/players/%d/snapshots"

func TestListPlayerPlayerSnapshots(t *testing.T) {
	t.Parallel()

	handler := newAPIHTTPHandler(t)
	pss := getAllPlayerSnapshots(t, handler)
	var server serverWithVersion
	var player apimodel.PlayerMeta

	pssGroupedByPlayerIDAndServerKey := make(map[string][]playerSnapshotWithServer)
	for _, ps := range pss {
		key := fmt.Sprintf("%d-%s", ps.Player.Id, ps.Server.Key)
		pssGroupedByPlayerIDAndServerKey[key] = append(pssGroupedByPlayerIDAndServerKey[key], ps)
	}
	currentMax := -1
	for _, grouped := range pssGroupedByPlayerIDAndServerKey {
		if l := len(grouped); l > currentMax && l > 0 {
			currentMax = l
			server = grouped[0].Server
			player = grouped[0].Player
		}
	}

	tests := []struct {
		name        string
		reqModifier func(t *testing.T, req *http.Request)
		assertResp  func(t *testing.T, req *http.Request, resp *http.Response)
	}{
		{
			name: "OK: without params",
			assertResp: func(t *testing.T, _ *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusOK, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ListPlayerSnapshotsResponse](t, resp.Body)
				assert.Zero(t, body.Cursor.Next)
				assert.NotZero(t, body.Cursor.Self)
				assert.NotZero(t, body.Data)
				assert.True(t, slices.IsSortedFunc(body.Data, func(a, b apimodel.PlayerSnapshot) int {
					return cmp.Or(
						a.Date.Compare(b.Date.Time),
						cmp.Compare(a.Id, b.Id),
					)
				}))
			},
		},
		{
			name: "OK: limit=1",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("limit", "1")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusOK, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ListPlayerSnapshotsResponse](t, resp.Body)
				assert.NotZero(t, body.Cursor.Next)
				assert.NotZero(t, body.Cursor.Self)
				assert.NotZero(t, body.Data)
				limit, err := strconv.Atoi(req.URL.Query().Get("limit"))
				require.NoError(t, err)
				assert.Len(t, body.Data, limit)
			},
		},
		{
			name: "OK: limit=1 cursor",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()

				q := req.URL.Query()
				q.Set("limit", "1")
				req.URL.RawQuery = q.Encode()

				resp := doCustomRequest(handler, req.Clone(req.Context()))
				defer resp.Body.Close()
				body := decodeJSON[apimodel.ListPlayerSnapshotsResponse](t, resp.Body)
				require.NotEmpty(t, body.Cursor.Next)

				q.Set("cursor", body.Cursor.Next)
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusOK, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ListPlayerSnapshotsResponse](t, resp.Body)
				assert.Equal(t, req.URL.Query().Get("cursor"), body.Cursor.Self)
				limit, err := strconv.Atoi(req.URL.Query().Get("limit"))
				require.NoError(t, err)
				assert.Len(t, body.Data, limit)
			},
		},
		{
			name: "OK: sort=[date:DESC]",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("sort", "date:DESC")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, _ *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusOK, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ListPlayerSnapshotsResponse](t, resp.Body)
				assert.Zero(t, body.Cursor.Next)
				assert.NotZero(t, body.Cursor.Self)
				assert.NotZero(t, body.Data)
				assert.True(t, slices.IsSortedFunc(body.Data, func(a, b apimodel.PlayerSnapshot) int {
					return cmp.Or(
						a.Date.Compare(b.Date.Time)*-1,
						cmp.Compare(a.Id, b.Id),
					)
				}))
			},
		},
		{
			name: "ERR: limit is not an integer",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("limit", "asd")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code: "invalid-param-format",
							Message: fmt.Sprintf(
								"error binding string parameter: strconv.ParseInt: parsing \"%s\": invalid syntax",
								req.URL.Query().Get("limit"),
							),
							Path: []string{"$query", "limit"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: limit < 1",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("limit", "0")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				limit, err := strconv.Atoi(req.URL.Query().Get("limit"))
				require.NoError(t, err)
				domainErr := domain.MinGreaterEqualError{
					Min:     1,
					Current: limit,
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"current": float64(domainErr.Current),
								"min":     float64(domainErr.Min),
							},
							Path: []string{"$query", "limit"},
						},
					},
				}, body)
			},
		},
		{
			name: fmt.Sprintf("ERR: limit > %d", domain.PlayerSnapshotListMaxLimit),
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("limit", strconv.Itoa(domain.PlayerSnapshotListMaxLimit+1))
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				limit, err := strconv.Atoi(req.URL.Query().Get("limit"))
				require.NoError(t, err)
				domainErr := domain.MaxLessEqualError{
					Max:     domain.PlayerSnapshotListMaxLimit,
					Current: limit,
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"current": float64(domainErr.Current),
								"max":     float64(domainErr.Max),
							},
							Path: []string{"$query", "limit"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: len(cursor) < 1",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("cursor", "")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				domainErr := domain.LenOutOfRangeError{
					Min:     1,
					Max:     1000,
					Current: len(req.URL.Query().Get("cursor")),
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"current": float64(domainErr.Current),
								"max":     float64(domainErr.Max),
								"min":     float64(domainErr.Min),
							},
							Path: []string{"$query", "cursor"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: len(cursor) > 1000",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("cursor", gofakeit.LetterN(1001))
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				domainErr := domain.LenOutOfRangeError{
					Min:     1,
					Max:     1000,
					Current: len(req.URL.Query().Get("cursor")),
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"current": float64(domainErr.Current),
								"max":     float64(domainErr.Max),
								"min":     float64(domainErr.Min),
							},
							Path: []string{"$query", "cursor"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: invalid cursor",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Set("cursor", gofakeit.LetterN(100))
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, _ *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				var domainErr domain.Error
				require.ErrorAs(t, domain.ErrInvalidCursor, &domainErr)
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Path:    []string{"$query", "cursor"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: len(sort) > 1",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Add("sort", "date:DESC")
				q.Add("sort", "date:ASC")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				domainErr := domain.LenOutOfRangeError{
					Min:     0,
					Max:     1,
					Current: len(req.URL.Query()["sort"]),
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"current": float64(domainErr.Current),
								"max":     float64(domainErr.Max),
								"min":     float64(domainErr.Min),
							},
							Path: []string{"$query", "sort"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: invalid sort",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				q := req.URL.Query()
				q.Add("sort", "date:")
				req.URL.RawQuery = q.Encode()
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusBadRequest, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				domainErr := domain.UnsupportedSortStringError{
					Sort: req.URL.Query()["sort"][0],
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"sort": domainErr.Sort,
							},
							Path: []string{"$query", "sort", "0"},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: version not found",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				req.URL.Path = fmt.Sprintf(
					endpointListPlayerPlayerSnapshots,
					randInvalidVersionCode(t, handler),
					server.Key,
					player.Id,
				)
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusNotFound, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				pathSegments := strings.Split(req.URL.Path, "/")
				require.Len(t, pathSegments, 9)
				domainErr := domain.VersionNotFoundError{
					VersionCode: pathSegments[3],
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"code": domainErr.VersionCode,
							},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: server not found",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				req.URL.Path = fmt.Sprintf(
					endpointListPlayerPlayerSnapshots,
					server.Version.Code,
					domaintest.RandServerKey(),
					player.Id,
				)
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusNotFound, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				pathSegments := strings.Split(req.URL.Path, "/")
				require.Len(t, pathSegments, 9)
				domainErr := domain.ServerNotFoundError{
					Key: pathSegments[5],
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"key": domainErr.Key,
							},
						},
					},
				}, body)
			},
		},
		{
			name: "ERR: player not found",
			reqModifier: func(t *testing.T, req *http.Request) {
				t.Helper()
				req.URL.Path = fmt.Sprintf(
					endpointListPlayerPlayerSnapshots,
					server.Version.Code,
					server.Key,
					domaintest.RandID(),
				)
			},
			assertResp: func(t *testing.T, req *http.Request, resp *http.Response) {
				t.Helper()

				assert.Equal(t, http.StatusNotFound, resp.StatusCode)

				// body
				body := decodeJSON[apimodel.ErrorResponse](t, resp.Body)
				pathSegments := strings.Split(req.URL.Path, "/")
				require.Len(t, pathSegments, 9)
				id, err := strconv.Atoi(pathSegments[7])
				require.NoError(t, err)
				domainErr := domain.PlayerNotFoundError{
					ID:        id,
					ServerKey: pathSegments[5],
				}
				assert.Equal(t, apimodel.ErrorResponse{
					Errors: []apimodel.Error{
						{
							Code:    apimodel.ErrorCode(domainErr.Code()),
							Message: domainErr.Error(),
							Params: map[string]any{
								"id":        float64(domainErr.ID),
								"serverKey": domainErr.ServerKey,
							},
						},
					},
				}, body)
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			req := httptest.NewRequest(
				http.MethodGet,
				fmt.Sprintf(endpointListPlayerPlayerSnapshots, server.Version.Code, server.Key, player.Id),
				nil,
			)
			if tt.reqModifier != nil {
				tt.reqModifier(t, req)
			}

			resp := doCustomRequest(handler, req)
			defer resp.Body.Close()
			tt.assertResp(t, req, resp)
		})
	}
}

type playerSnapshotWithServer struct {
	apimodel.PlayerSnapshot
	Server serverWithVersion
}

func getAllPlayerSnapshots(tb testing.TB, h http.Handler) []playerSnapshotWithServer {
	tb.Helper()

	players := getAllPlayers(tb, h)

	var pss []playerSnapshotWithServer

	for _, p := range players {
		resp := doRequest(h, http.MethodGet, fmt.Sprintf(
			endpointListPlayerPlayerSnapshots,
			p.Server.Version.Code,
			p.Server.Key,
			p.Id,
		), nil)
		require.Equal(tb, http.StatusOK, resp.StatusCode)

		for _, ps := range decodeJSON[apimodel.ListPlayerSnapshotsResponse](tb, resp.Body).Data {
			pss = append(pss, playerSnapshotWithServer{
				PlayerSnapshot: ps,
				Server:         p.Server,
			})
		}

		_ = resp.Body.Close()
	}

	require.NotZero(tb, pss)

	return pss
}
